@@ -8,62 +8,79 @@ from sql.aggregate import Max, Count, Sum
 from sql.conditionals import Case, Coalesce
 from sql.functions import Extract
 
+from trytond.i18n import gettext
 from trytond.model import ModelView, ModelSQL, Workflow, fields, \
     sequence_ordered
+from trytond.model.exceptions import AccessError
 from trytond.pyson import Eval, In, If, Get, Bool
 from trytond.transaction import Transaction
 from trytond.pool import Pool
 
-__all__ = ['SaleOpportunity', 'SaleOpportunityLine',
-    'SaleOpportunityEmployee', 'SaleOpportunityEmployeeContext',
-    'SaleOpportunityMonthly', 'SaleOpportunityEmployeeMonthly']
-
-STATES = [
-    ('lead', 'Lead'),
-    ('opportunity', 'Opportunity'),
-    ('converted', 'Converted'),
-    ('won', 'Won'),
-    ('cancelled', 'Cancelled'),
-    ('lost', 'Lost'),
-]
-_STATES_START = {
-    'readonly': Eval('state') != 'lead',
-    }
-_DEPENDS_START = ['state']
-_STATES_STOP = {
-    'readonly': In(Eval('state'), ['converted', 'won', 'lost', 'cancelled']),
-}
-_DEPENDS_STOP = ['state']
+from trytond.ir.attachment import AttachmentCopyMixin
+from trytond.ir.note import NoteCopyMixin
+from trytond.modules.company.model import employee_field, set_employee
 
 
-class SaleOpportunity(Workflow, ModelSQL, ModelView):
+class SaleOpportunity(
+        Workflow, ModelSQL, ModelView,
+        AttachmentCopyMixin, NoteCopyMixin):
     'Sale Opportunity'
     __name__ = "sale.opportunity"
     _history = True
     _rec_name = 'number'
+
+    _states_start = {
+        'readonly': Eval('state') != 'lead',
+        }
+    _depends_start = ['state']
+    _states_stop = {
+        'readonly': Eval('state').in_(
+            ['converted', 'won', 'lost', 'cancelled']),
+    }
+    _depends_stop = ['state']
+
     number = fields.Char('Number', readonly=True, required=True, select=True)
     reference = fields.Char('Reference', select=True)
-    party = fields.Many2One('party.party', 'Party', select=True,
+    party = fields.Many2One(
+        'party.party', "Party", select=True,
         states={
             'readonly': Eval('state').in_(['converted', 'lost', 'cancelled']),
             'required': ~Eval('state').in_(['lead', 'lost', 'cancelled']),
-            }, depends=['state'])
+            },
+        context={
+            'company': Eval('company', -1),
+            },
+        depends=['state', 'company'])
+    contact = fields.Many2One(
+        'party.contact_mechanism', "Contact",
+        context={
+            'company': Eval('company', -1),
+            },
+        search_context={
+            'related_party': Eval('party'),
+            },
+        depends=['party', 'company'])
     address = fields.Many2One('party.address', 'Address',
         domain=[('party', '=', Eval('party'))],
         select=True, depends=['party', 'state'],
-        states=_STATES_STOP)
+        states=_states_stop)
     company = fields.Many2One('company.company', 'Company', required=True,
-        select=True, states=_STATES_STOP, domain=[
+        select=True,
+        states={
+            'readonly': _states_stop['readonly'] | Eval('party', True),
+            },
+        domain=[
             ('id', If(In('company', Eval('context', {})), '=', '!='),
                 Get(Eval('context', {}), 'company', 0)),
-            ], depends=_DEPENDS_STOP)
+            ],
+        depends=_depends_stop)
     currency = fields.Function(fields.Many2One('currency.currency',
         'Currency'), 'get_currency')
     currency_digits = fields.Function(fields.Integer('Currency Digits'),
             'get_currency_digits')
     amount = fields.Numeric('Amount', digits=(16, Eval('currency_digits', 2)),
-        states=_STATES_STOP, depends=_DEPENDS_STOP + ['currency_digits'],
-        help='Estimated revenue amount')
+        states=_states_stop, depends=_depends_stop + ['currency_digits'],
+        help='Estimated revenue amount.')
     payment_term = fields.Many2One('account.invoice.payment_term',
         'Payment Term', states={
             'readonly': In(Eval('state'),
@@ -72,23 +89,21 @@ class SaleOpportunity(Workflow, ModelSQL, ModelView):
         depends=['state'])
     employee = fields.Many2One('company.employee', 'Employee',
         states={
-            'readonly': _STATES_STOP['readonly'],
+            'readonly': _states_stop['readonly'],
             'required': ~Eval('state').in_(['lead', 'lost', 'cancelled']),
         },
         depends=['state', 'company'],
         domain=[('company', '=', Eval('company'))])
     start_date = fields.Date('Start Date', required=True, select=True,
-        states=_STATES_START, depends=_DEPENDS_START)
+        states=_states_start, depends=_depends_start)
     end_date = fields.Date('End Date', select=True,
-        states=_STATES_STOP, depends=_DEPENDS_STOP)
-    description = fields.Char('Description', required=True,
-        states=_STATES_STOP, depends=_DEPENDS_STOP)
-    comment = fields.Text('Comment', states=_STATES_STOP,
-        depends=_DEPENDS_STOP)
+        states=_states_stop, depends=_depends_stop)
+    description = fields.Char('Description',
+        states=_states_stop, depends=_depends_stop)
+    comment = fields.Text('Comment', states=_states_stop,
+        depends=_depends_stop)
     lines = fields.One2Many('sale.opportunity.line', 'opportunity', 'Lines',
-        states=_STATES_STOP, depends=_DEPENDS_STOP)
-    state = fields.Selection(STATES, 'State', required=True, select=True,
-            sort=False, readonly=True)
+        states=_states_stop, depends=_depends_stop)
     conversion_probability = fields.Float('Conversion Probability',
         digits=(1, 4), required=True,
         domain=[
@@ -99,17 +114,34 @@ class SaleOpportunity(Workflow, ModelSQL, ModelView):
             'readonly': ~Eval('state').in_(
                 ['opportunity', 'lead', 'converted']),
             },
-        depends=['state'], help="Percentage between 0 and 100")
+        depends=['state'], help="Percentage between 0 and 100.")
     lost_reason = fields.Text('Reason for loss', states={
             'invisible': Eval('state') != 'lost',
             }, depends=['state'])
     sales = fields.One2Many('sale.sale', 'origin', 'Sales')
 
+    converted_by = employee_field(
+        "Converted By", states=['converted', 'won', 'lost', 'cancelled'])
+    state = fields.Selection([
+            ('lead', "Lead"),
+            ('opportunity', "Opportunity"),
+            ('converted', "Converted"),
+            ('won', "Won"),
+            ('cancelled', "Cancelled"),
+            ('lost', "Lost"),
+            ], "State", required=True, select=True,
+        sort=False, readonly=True)
+
+    del _states_start, _depends_start
+    del _states_stop, _depends_stop
+
     @classmethod
     def __register__(cls, module_name):
         pool = Pool()
         Sale = pool.get('sale.sale')
-        cursor = Transaction().connection.cursor()
+        transaction = Transaction()
+        cursor = transaction.connection.cursor()
+        update = transaction.connection.cursor()
         sql_table = cls.__table__()
         sale = Sale.__table__()
 
@@ -129,8 +161,8 @@ class SaleOpportunity(Workflow, ModelSQL, ModelView):
             cursor.execute(*sql_table.select(
                     sql_table.id, sql_table.sale,
                     where=sql_table.sale != Null))
-            for id_, sale_id in cursor.fetchall():
-                cursor.execute(*sale.update(
+            for id_, sale_id in cursor:
+                update.execute(*sale.update(
                         columns=[sale.origin],
                         values=['%s,%s' % (cls.__name__, id_)],
                         where=sale.id == sale_id))
@@ -147,14 +179,13 @@ class SaleOpportunity(Workflow, ModelSQL, ModelView):
         # Migration from 4.2: make employee not required
         table.not_null_action('employee', action='remove')
 
+        # Migration from 5.0: drop required on description
+        table.not_null_action('description', action='remove')
+
     @classmethod
     def __setup__(cls):
         super(SaleOpportunity, cls).__setup__()
         cls._order.insert(0, ('start_date', 'DESC'))
-        cls._error_messages.update({
-                'delete_cancel': ('Sale Opportunity "%s" must be cancelled '
-                    'before deletion.'),
-                })
         cls._transitions |= set((
                 ('lead', 'opportunity'),
                 ('lead', 'lost'),
@@ -225,18 +256,31 @@ class SaleOpportunity(Workflow, ModelSQL, ModelView):
         if len(payment_terms) == 1:
             return payment_terms[0].id
 
+    @classmethod
+    def view_attributes(cls):
+        return super().view_attributes() + [
+            ('/tree', 'visual', If(Eval('state') == 'cancelled', 'muted', '')),
+            ]
+
+    @classmethod
+    def get_resources_to_copy(cls, name):
+        return {
+            'sale.sale',
+            }
+
     @classmethod
     def create(cls, vlist):
         pool = Pool()
-        Sequence = pool.get('ir.sequence')
         Config = pool.get('sale.configuration')
 
         config = Config(1)
         vlist = [x.copy() for x in vlist]
+        default_company = cls.default_company()
         for vals in vlist:
             if vals.get('number') is None:
-                vals['number'] = Sequence.get_id(
-                    config.sale_opportunity_sequence.id)
+                vals['number'] = config.get_multivalue(
+                    'sale_opportunity_sequence',
+                    company=vals.get('company', default_company)).get()
         return super(SaleOpportunity, cls).create(vlist)
 
     @classmethod
@@ -247,6 +291,7 @@ class SaleOpportunity(Workflow, ModelSQL, ModelView):
             default = default.copy()
         default.setdefault('number', None)
         default.setdefault('sales', None)
+        default.setdefault('converted_by')
         return super(SaleOpportunity, cls).copy(opportunities, default=default)
 
     def get_currency(self, name):
@@ -276,6 +321,7 @@ class SaleOpportunity(Workflow, ModelSQL, ModelView):
         return Sale(
             description=self.description,
             party=self.party,
+            contact=self.contact,
             payment_term=self.payment_term,
             company=self.company,
             invoice_address=self.address,
@@ -304,7 +350,9 @@ class SaleOpportunity(Workflow, ModelSQL, ModelView):
         cls.cancel(opportunities)
         for opportunity in opportunities:
             if opportunity.state != 'cancelled':
-                cls.raise_user_error('delete_cancel', opportunity.rec_name)
+                raise AccessError(
+                    gettext('sale_opportunity.msg_opportunity_delete_cancel',
+                        opportunity=opportunity.rec_name))
         super(SaleOpportunity, cls).delete(opportunities)
 
     @classmethod
@@ -322,11 +370,14 @@ class SaleOpportunity(Workflow, ModelSQL, ModelView):
     @classmethod
     @ModelView.button
     @Workflow.transition('converted')
+    @set_employee('converted_by')
     def convert(cls, opportunities):
         pool = Pool()
         Sale = pool.get('sale.sale')
         sales = [o.create_sale() for o in opportunities if not o.sales]
         Sale.save(sales)
+        for sale in sales:
+            sale.origin.copy_resources_to(sale)
 
     @property
     def is_forecast(self):
@@ -370,7 +421,7 @@ class SaleOpportunity(Workflow, ModelSQL, ModelView):
 
     @staticmethod
     def _sale_lost_states():
-        return ['cancel']
+        return ['cancelled']
 
     def is_won(self):
         sale_won_states = self._sale_won_states()
@@ -437,12 +488,13 @@ class SaleOpportunityLine(sequence_ordered(), ModelSQL, ModelView):
     _depends = ['opportunity_state']
 
     opportunity = fields.Many2One('sale.opportunity', 'Opportunity',
+        ondelete='CASCADE', select=True, required=True,
         states={
             'readonly': _states['readonly'] & Bool(Eval('opportunity')),
             },
         depends=_depends)
     opportunity_state = fields.Function(
-        fields.Selection(STATES, 'Opportunity State'),
+        fields.Selection('get_opportunity_states', "Opportunity State"),
         'on_change_with_opportunity_state')
     product = fields.Many2One('product.product', 'Product', required=True,
         domain=[('salable', '=', True)], states=_states, depends=_depends)
@@ -456,6 +508,17 @@ class SaleOpportunityLine(sequence_ordered(), ModelSQL, ModelView):
 
     del _states, _depends
 
+    @classmethod
+    def __setup__(cls):
+        super().__setup__()
+        cls.__access__.add('opportunity')
+
+    @classmethod
+    def get_opportunity_states(cls):
+        pool = Pool()
+        Opportunity = pool.get('sale.opportunity')
+        return Opportunity.fields_get(['state'])['state']['selection']
+
     @fields.depends('opportunity', '_parent_opportunity.state')
     def on_change_with_opportunity_state(self, name=None):
         if self.opportunity:
@@ -498,7 +561,14 @@ class SaleOpportunityLine(sequence_ordered(), ModelSQL, ModelView):
         sale_line.unit = self.unit
 
     def get_rec_name(self, name):
-        return self.product.rec_name
+        pool = Pool()
+        Lang = pool.get('ir.lang')
+        lang = Lang.get()
+        return (lang.format(
+                '%.*f', (self.unit.digits, self.quantity or 0))
+            + '%s %s @ %s' % (
+                self.unit.symbol, self.product.rec_name,
+                self.opportunity.rec_name))
 
     @classmethod
     def search_rec_name(cls, name, clause):
@@ -506,6 +576,7 @@ class SaleOpportunityLine(sequence_ordered(), ModelSQL, ModelView):
 
 
 class SaleOpportunityReportMixin:
+    __slots__ = ()
     number = fields.Integer('Number')
     converted = fields.Integer('Converted')
     conversion_rate = fields.Function(fields.Float('Conversion Rate',
@@ -590,17 +661,22 @@ class SaleOpportunityReportMixin:
             Max(opportunity.write_date).as_('write_date'),
             opportunity.company,
             Count(Literal(1)).as_('number'),
-            Sum(Case((opportunity.state.in_(cls._converted_state()),
+            Sum(Case(
+                    (opportunity.state.in_(cls._converted_state()),
                         Literal(1)), else_=Literal(0))).as_('converted'),
-            Sum(Case((opportunity.state.in_(cls._won_state()),
+            Sum(Case(
+                    (opportunity.state.in_(cls._won_state()),
                         Literal(1)), else_=Literal(0))).as_('won'),
-            Sum(Case((opportunity.state.in_(cls._lost_state()),
+            Sum(Case(
+                    (opportunity.state.in_(cls._lost_state()),
                         Literal(1)), else_=Literal(0))).as_('lost'),
             Sum(opportunity.amount).as_('amount'),
-            Sum(Case((opportunity.state.in_(cls._converted_state()),
+            Sum(Case(
+                    (opportunity.state.in_(cls._converted_state()),
                         opportunity.amount),
                     else_=Literal(0))).as_('converted_amount'),
-            Sum(Case((opportunity.state.in_(cls._won_state()),
+            Sum(Case(
+                    (opportunity.state.in_(cls._won_state()),
                         opportunity.amount),
                     else_=Literal(0))).as_('won_amount'))
 
@@ -620,11 +696,11 @@ class SaleOpportunityEmployee(SaleOpportunityReportMixin, ModelSQL, ModelView):
             )
         where = Literal(True)
         if Transaction().context.get('start_date'):
-            where &= (opportunity.start_date >=
-                Transaction().context['start_date'])
+            where &= (opportunity.start_date
+                >= Transaction().context['start_date'])
         if Transaction().context.get('end_date'):
-            where &= (opportunity.start_date <=
-                Transaction().context['end_date'])
+            where &= (opportunity.start_date
+                <= Transaction().context['end_date'])
         query.where = where
         query.group_by = (opportunity.employee, opportunity.company)
         return query
@@ -641,7 +717,7 @@ class SaleOpportunityMonthly(SaleOpportunityReportMixin, ModelSQL, ModelView):
     'Sale Opportunity per Month'
     __name__ = 'sale.opportunity_monthly'
     year = fields.Char('Year')
-    month = fields.Integer('Month')
+    month = fields.Many2One('ir.calendar.month', "Month")
     year_month = fields.Function(fields.Char('Year-Month'),
             'get_year_month')
 
@@ -649,27 +725,32 @@ class SaleOpportunityMonthly(SaleOpportunityReportMixin, ModelSQL, ModelView):
     def __setup__(cls):
         super(SaleOpportunityMonthly, cls).__setup__()
         cls._order.insert(0, ('year', 'DESC'))
-        cls._order.insert(1, ('month', 'DESC'))
+        cls._order.insert(1, ('month.index', 'DESC'))
 
     def get_year_month(self, name):
-        return '%s-%s' % (self.year, int(self.month))
+        return '%s-%s' % (self.year, self.month.index)
 
     @classmethod
     def table_query(cls):
+        pool = Pool()
+        Month = pool.get('ir.calendar.month')
+        month = Month.__table__()
         query = super(SaleOpportunityMonthly, cls).table_query()
         opportunity, = query.from_
         type_id = cls.id.sql_type().base
         type_year = cls.year.sql_type().base
         year_column = Extract('YEAR', opportunity.start_date
             ).cast(type_year).as_('year')
-        month_column = Extract('MONTH', opportunity.start_date).as_('month')
+        month_index = Extract('MONTH', opportunity.start_date)
+        query.from_ = opportunity.join(
+            month, condition=month_index == month.index)
         query.columns += (
             Max(Extract('MONTH', opportunity.start_date)
                 + Extract('YEAR', opportunity.start_date) * 100
                 ).cast(type_id).as_('id'),
             year_column,
-            month_column)
-        query.group_by = (year_column, month_column, opportunity.company)
+            month.id.as_('month'))
+        query.group_by = (year_column, month.id, opportunity.company)
         return query
 
 
@@ -678,34 +759,39 @@ class SaleOpportunityEmployeeMonthly(
     'Sale Opportunity per Employee per Month'
     __name__ = 'sale.opportunity_employee_monthly'
     year = fields.Char('Year')
-    month = fields.Integer('Month')
+    month = fields.Many2One('ir.calendar.month', "Month")
     employee = fields.Many2One('company.employee', 'Employee')
 
     @classmethod
     def __setup__(cls):
         super(SaleOpportunityEmployeeMonthly, cls).__setup__()
         cls._order.insert(0, ('year', 'DESC'))
-        cls._order.insert(1, ('month', 'DESC'))
+        cls._order.insert(1, ('month.index', 'DESC'))
         cls._order.insert(2, ('employee', 'ASC'))
 
     @classmethod
     def table_query(cls):
+        pool = Pool()
+        Month = pool.get('ir.calendar.month')
+        month = Month.__table__()
         query = super(SaleOpportunityEmployeeMonthly, cls).table_query()
         opportunity, = query.from_
         type_id = cls.id.sql_type().base
         type_year = cls.year.sql_type().base
         year_column = Extract('YEAR', opportunity.start_date
             ).cast(type_year).as_('year')
-        month_column = Extract('MONTH', opportunity.start_date).as_('month')
+        month_index = Extract('MONTH', opportunity.start_date)
+        query.from_ = opportunity.join(
+            month, condition=month_index == month.index)
         query.columns += (
             Max(Extract('MONTH', opportunity.start_date)
                 + Extract('YEAR', opportunity.start_date) * 100
                 + Coalesce(opportunity.employee, 0) * 1000000
                 ).cast(type_id).as_('id'),
             year_column,
-            month_column,
+            month.id.as_('month'),
             opportunity.employee,
             )
-        query.group_by = (year_column, month_column,
+        query.group_by = (year_column, month.id,
             opportunity.employee, opportunity.company)
         return query
